#include <stdio.h>

/* Private data structure */
typedef struct { int private; } SBG;

/**
 * Initialise a SBaGen engine
 */
extern SBG *sbg_init(double sample_freq);

/**
 * Free a SBaGen engine and all associated resources.
 */
extern void sbg_term(SBG *sbg);

/**
 * Return the library version as a string.
 */
extern char *sbg_version();

/**
 * Generate a segment of 16-bit dithered audio from the sbagen engine.
 * This advances the 'now' time.  Writes 'cnt' 16-bit stereo samples
 * to the given buffer.  This does no allocations or deallocations, so
 * can be called from time-sensitive code (e.g. audio callback).  It
 * can be called from another thread if necessary.
 */
extern void sbg_gen(SBG *sbg, short *buf, int outcnt);

/**
 * Generate a segment of 32-bit audio from the sbagen engine without
 * dithering.  Note that the SBaGen library is designed for 16 bit
 * output, so there is no guaranteed fidelity gain using this method,
 * although for some types of generation (e.g. MP3 decoding or
 * low-amplitude binaural beats) this will give greater fidelity.
 *
 * The call advances the 'now' time.  Writes 'outcnt' 32-bit stereo
 * samples to the given buffer.  This does no allocations or
 * deallocations, so can be called from time-sensitive code.  It can
 * be called from another thread if necessary.
 */
extern void sbg_gen32(SBG *sbg, int *buf, int outcnt);

/**
 * Return the current time in samples.  This can be used to add live
 * events with sbg_set() to make things happen at a time relative to
 * now.
 */
extern int sbg_now(SBG *sbg);

/**
 * Return the sample rate that this engine runs at (as passed to
 * sbg_init()).
 */
extern double sbg_rate(SBG* sbg);

/**
 * Add a new spec to one of the channels.  Each channel keeps a list
 * of specs which it works through as time advances.  This call adds a
 * new spec.  If 'time' is earlier than one or more existing specs,
 * then those specs are deleted and replaced by the new spec.  If
 * 'slideto' is true, then a slide is made from the previous spec to
 * this new spec (if possible).  If the previous spec is the one
 * currently playing, then the slide starts immediately.
 *
 * Specification of "-" is special and means 'off'.  Slide to or from
 * 'off' is supported and adjusts its meaning according to the other
 * side of the slide.
 *
 * If 'spec' is invalid, then a strdup'd error string is returned
 * (which the caller must free).  On success, NULL is returned.
 */
extern char *sbg_set(SBG *sbg, int chan, int time, int slideto, char *spec);

/**
 * Dump the current state and all the currently loaded events to the
 * given stream for debugging.
 */
extern void sbg_debug_dump(SBG *sbg, FILE *out);

/**
 * Start outputting audio from the given SBG engine using PortAudio.
 * The time 'now' will start advancing immediately.  This is suited to
 * playing sequences that have already been loaded up into the engine,
 * or for playing real-time sequences which are generated in real time
 * relative to the time 'now'.  Note that 'now' will be ahead of
 * actual time according to the latency introduced by the audio
 * buffers, which depends on the platform.
 *
 * Returns NULL on success, or otherwise a strdup'd error string which
 * the caller must free.
 */
extern char *sbg_pa_start(SBG *sbg);

/**
 * Stop playing and close the PortAudio device.  The SBG engine will
 * freeze with whatever value 'now' had before this call.  If playing
 * is restarted later, it will continue from this point.
 *
 * Returns NULL on success, or otherwise a strdup'd error string which
 * the caller must free.
 */
extern char *sbg_pa_stop();

