//
//	Sequenced Binaural Beat Generator Library
//
//        Copyright (c) 1999-2011 Jim Peters <http://uazu.net/>.
//        Released under the GNU GPL version 2 as published by the
//        Free Software Foundation.  See the file COPYING for details,
//        or visit <http://www.gnu.org/copyleft/gpl.html>.
//
//	Note that this file is released under the *GNU GPL*, so this
//	means that it can only be linked with code that uses a
//	compatible open-source / free software licence.  For an
//	overview of compatible and incompatible licences, see here:
//
//	  http://www.fsf.org/licenses/license-list.html
//
//	If you want to use this code in a project that can't be put
//	under a compatible licence (e.g. closed-source), then you
//	need specific permission from me to use this code.
//
//	-- Jim Peters   (http://uazu.net/contact/)
//

#define VERSION "2.0.1"

#define ALLOC_ARR(cnt, type) ((type*)Alloc((cnt) * sizeof(type)))
#define ALLOC(type) ((type*)Alloc(sizeof(type)))

// Time comparison macros (that work even when times wrap)
#define T_GT(a,b)  (((a)-(b)) > 0)
#define T_LT(a,b)  (((a)-(b)) < 0)
#define T_GE(a,b)  (((a)-(b)) >= 0)
#define T_LE(a,b)  (((a)-(b)) <= 0)

typedef unsigned int U32;
typedef signed int S32;
typedef unsigned short U16;
typedef signed short S16;

typedef struct SBG SBG;
typedef struct SBG_Chan SBG_Chan;
typedef struct SBG_Voice SBG_Voice;
typedef struct SBG_VTyp SBG_VTyp;

// Headers
#include <stdio.h>
#include <stdlib.h>
#include <stdarg.h>
#include <string.h>
#include <math.h>
#include <ctype.h>

// Structures
struct SBG {
   SBG_Chan *chans;   // List of active channels
   double rate;       // Current sampling rate
   S32 now;           // Current output sample count (wraps)
   S32 *genbuf;       // Internal buffer used for generating audio
   int genbuf_cnt;    // Number of 32-bit stereo samples in 'genbuf'
};

struct SBG_Chan {
   SBG_Chan *nxt;  // Next or NULL
   int chan;       // Channel number
   SBG_Voice *voice;   // List of current voices, or NULL
   SBG_Voice *curr;    // Used by 'gen': points to currently outputting voice, or NULL
};

struct SBG_Voice {
   SBG_Voice *nxt;    // Next voice in time sequence, or NULL
   S32 time;          // Time that this voice becomes active
   int is_off;        // Is this an 'off' voice?
   SBG_VTyp *vtyp;    // Type of this voice definition
};

// Voice type
struct SBG_VTyp {
   /**
    * Create a new voice if the spec parses correctly for this type,
    * else return NULL.  On failure, sets *err to NULL if this is a
    * bad match to this type, or to a strdup'd error string if it
    * matches but parses badly or has other problems (e.g. files not
    * found, etc).  Note that "-" should always be accepted as the
    * 'off' version of this voice.
    */
   SBG_Voice* (*init)(SBG *sbg, char *spec, char **err);

   /**
    * Duplicate this voice with amplitude 0.  This is used to create a
    * slide from off to an existing voice setting.
    */
   SBG_Voice* (*offdup)(SBG *sbg, SBG_Voice *v1);
   
   /**
    * Release resources and 'free' the voice.
    */
   void (*term)(SBG_Voice*);

   /**
    * Set up a slide from one voice to another (of same type), or
    * cancel slide if 'v2' is NULL.
    */
   void (*slide)(SBG_Voice* v1, SBG_Voice* v2);

   /**
    * Transfer phase information forwards from one voice to the next
    * (of the same type).
    */
   void (*xfer)(SBG_Voice* v1, SBG_Voice* v2);

   /**
    * Generate output, combining 'cnt' stereo samples (S32,S32) with
    * data already at *dst++.  Voice being output is 'v1'.  This is a
    * time-critical routine, so it should do no memory allocation or
    * deallocation, and no disk accesses.
    */
   void (*gen)(SBG_Voice* v1, S32 cnt, S32 *dst);

   /**
    * Dump current voice information to 'out' (with newline)
    */
   void (*dump)(SBG_Voice* vv, FILE *out);
};

// Utility routines
#include "util.c"
#include "parse.c"
   
// Voice generators
#include "voice_bin.c"

//
//	Voice types
//

static SBG_VTyp voice_type[]= {
   // Binaural generator
   { bin_init, bin_offdup, bin_term, bin_slide, bin_xfer, bin_gen, bin_dump },

   // END
   { NULL }
};


/**
 * Initialise a SBaGen engine
 */
SBG*
sbg_init(double sample_freq) {
   SBG *rv = ALLOC(SBG);
   rv->rate = sample_freq;
   rv->now = 0;

   {
      int cnt = (int) (sample_freq / 10);
      while ((cnt & (cnt-1)) != 0) cnt &= cnt-1;
      rv->genbuf_cnt = cnt;
      rv->genbuf = ALLOC_ARR(cnt * 2, S32);
   }

   init_sin_table();
   return rv;
}

/**
 * Free a SBaGen engine and all associated resources.
 */
void
sbg_term(SBG *sbg) {
   while (sbg->chans) {
      SBG_Chan *ch = sbg->chans;
      sbg->chans = ch->nxt;
      while (ch->voice) {
         SBG_Voice *vv = ch->voice;
         ch->voice = vv->nxt;
         vv->vtyp->term(vv);
      }
      free(ch);
   }
   free(sbg->genbuf);
   free(sbg);
}

/**
 * Return the library version as a string.
 */
char *
sbg_version() {
   return VERSION;
}

/**
 * Fill the given S32 buffer with the next 'cnt' stereo samples and
 * advance 'now'.  The buffer should already have been zero'd.
 */
static void generate(SBG *sbg, S32 *dst0, int cnt) {
   SBG_Chan *ch;
   for (ch = sbg->chans; ch; ch=ch->nxt) {
      S32 now = sbg->now;
      S32 end = now + cnt;
      SBG_Voice *v1 = ch->curr;
      if (!v1) continue;
      SBG_Voice *v2 = v1->nxt;
      while (v2 && T_LE(v2->time, now)) {
         if (v1->vtyp == v2->vtyp)
            v1->vtyp->xfer(v1, v2);
         v1 = ch->curr = v2;
         v2 = v1->nxt;
      }
      S32 *dst = dst0;
      while (T_LT(now, end)) {
         S32 n = end - now;
         if (v2 && T_LT(v2->time, end)) {
            n = v2->time - now;
         }
         if (n > 0) {
            v1->vtyp->gen(v1, n, dst);
            dst += n * 2;
            now += n;
         }
         if (v2 && T_LE(v2->time, now)) {
            if (v1->vtyp == v2->vtyp) 
               v1->vtyp->xfer(v1, v2);
            v1 = ch->curr = v2;
            v2 = v1->nxt;
         }
      }
   }
   
   // Advance 'cnt' samples
   sbg->now += cnt;
}

/**
 * White noise source for dithering.
 */
static S32 dither = 0;

/**
 * Generate a segment of 16-bit dithered audio from the sbagen engine.
 * This advances the 'now' time.  Writes 'outcnt' 16-bit stereo
 * samples to the given buffer.  This does no allocations or
 * deallocations, so can be called from time-sensitive code.  It can
 * be called from another thread if necessary.
 */
void
sbg_gen(SBG *sbg, S16 *buf, int outcnt) {
   while (outcnt > 0) {
      int cnt = sbg->genbuf_cnt;
      if (cnt > outcnt)
         cnt = outcnt;
      memset(sbg->genbuf, 0, cnt * 2 * sizeof(S32));

      // Fill buffer and advance 'cnt' samples
      generate(sbg, sbg->genbuf, cnt);
      outcnt -= cnt;

      // Convert data to 16-bit for output, adding dithering
      {
         int a;
         S32 *src = sbg->genbuf;
         
         for (a = 0; a<cnt; a++) {
            // White noise dither
            dither = (dither * 0x660D + 0xF35F) & 0xFFFF;

            // Values over 0x7FFF0000 may overflow
            *buf++ = (*src++ + dither) >> 16;
            *buf++ = (*src++ + dither) >> 16;
         }
      }
   }
}

/**
 * Generate a segment of 32-bit audio from the sbagen engine without
 * dithering.  Note that the SBaGen library is designed for 16 bit
 * output, so there is no guaranteed fidelity gain using this method,
 * although for some types of generation (e.g. MP3 decoding or
 * low-amplitude binaural beats) this will give greater fidelity.
 *
 * The call advances the 'now' time.  Writes 'outcnt' 32-bit stereo
 * samples to the given buffer.  This does no allocations or
 * deallocations, so can be called from time-sensitive code.  It can
 * be called from another thread if necessary.
 */
void
sbg_gen32(SBG *sbg, S32 *buf, int outcnt) {
   memset(buf, 0, outcnt * 2 * sizeof(S32));
   generate(sbg, buf, outcnt);
}

/**
 * Return the current time in samples.  This can be used to add live
 * events with sbg_set() to make things happen at a time relative to
 * now.
 */
S32
sbg_now(SBG* sbg) {
   return sbg->now;
}

/**
 * Return the sample rate that this engine runs at (as passed to
 * sbg_init()).
 */
double
sbg_rate(SBG* sbg) {
   return sbg->rate;
}

/**
 * Add a new spec to one of the channels.  Each channel keeps a list
 * of specs which it works through as time advances.  This call adds a
 * new spec.  If 'time' is earlier than one or more existing specs,
 * then those specs are deleted and replaced by the new spec.  If
 * 'slideto' is true, then a slide is made from the previous spec to
 * this new spec (if possible).  If the previous spec is the one
 * currently playing, then the slide starts immediately.
 *
 * Specification of "-" is special and means 'off'.  Slide to or from
 * 'off' is supported and adjusts its meaning according to the other
 * side of the slide.
 *
 * If 'spec' is invalid, then a strdup'd error string is returned
 * (which the caller must free).  On success, NULL is returned.
 */
char *
sbg_set(SBG* sbg, int chan, int time, int slideto, char *spec) {
   char *err;
   SBG_Chan *ch;
   SBG_Voice *v1 = NULL;
   SBG_Voice *vv;
   SBG_Voice *v2 = NULL;
   int is_off;

   // Find or create channel
   for (ch = sbg->chans; ch; ch = ch->nxt) {
      if (ch->chan == chan)
         break;
   }
   if (!ch) {
      ch = ALLOC(SBG_Chan);
      ch->chan = chan;
      ch->nxt = sbg->chans;
      sbg->chans = ch;
   }

   // Cleanup expired voices
   while (ch->voice != ch->curr) {
      SBG_Voice *freeme = ch->voice;
      ch->voice = freeme->nxt;
      freeme->vtyp->term(freeme);
   }

   // Look for which voice this new voice should follow
   v1 = NULL;
   for (vv = ch->voice; vv; vv = vv->nxt) {
      if (T_LT(vv->time, time))
         v1 = vv;
   }
   if (!v1) {
      v1 = voice_type[0].init(sbg, "-", &err);
      v1->vtyp = &voice_type[0];
      v1->time = sbg->now;
      v1->is_off = 1;
      ch->voice = v1;
      ch->curr = v1;
   }

   // Clear any later voices which have now been overridden by
   // this new voice
   while (v1 && v1->nxt) {
      SBG_Voice *freeme = v1->nxt;
      v1->nxt = freeme->nxt;
      freeme->vtyp->term(freeme);
   }

   // Parse the voice
   is_off = 0 == strcmp(spec, "-");
   if (is_off) {
      // "-" should always be accepted by init()
      if (slideto && v1) {
         v2 = v1->vtyp->init(sbg, "-", &err);
         v2->vtyp = v1->vtyp;
      } else {
         v2 = voice_type[0].init(sbg, "-", &err);
         v2->vtyp = &voice_type[0];
      }
   } else {
      int a;
      for (a = 0; voice_type[a].init; a++) {
         v2 = voice_type[a].init(sbg, spec, &err);
         if (!v2) {
            if (err) return err;
            continue;
         }
         v2->vtyp = &voice_type[a];
         break;
      }
      if (!v2)
         return StrDup("Spec does not match any known voice type");
   }

   // Add to list
   v2->is_off = is_off;
   v2->time = time;
   v1->nxt = v2;

   // Handle slides: first cancel any slide, then set one up if required
   if (v1) {
      v1->vtyp->slide(v1, NULL);
      if (slideto) {
         // If v1 is already playing, then start slide from 'now'
         if (T_LT(v1->time, sbg->now))
            v1->time = sbg->now;
         
         if (v1->is_off) {
            // Insert new 'off' voice with the right type
            vv = v2->vtyp->offdup(sbg, v2);
            vv->time = v1->time;
            vv->is_off = 1;
            vv->nxt = v2;
            v1->nxt = vv;
            v1 = vv;
         }
         if (v1->vtyp == v2->vtyp) {
            v1->vtyp->slide(v1, v2);
         }
      }
   }

   return NULL;
}

/**
 * Dump the current state and all the currently loaded events to the
 * given stream for debugging.
 */
void 
sbg_debug_dump(SBG *sbg, FILE *out) {
   SBG_Chan *ch;
   SBG_Voice *vc;
   
   fprintf(out, "Sampling rate: %lg\n", sbg->rate);
   fprintf(out, "Genbuf length: %d samples\n", sbg->genbuf_cnt);
   fprintf(out, "Now time:      %d ms\n", sbg->now);

   for (ch = sbg->chans; ch; ch = ch->nxt) {
      fprintf(out, "Channel %d:\n", ch->chan);
      for (vc = ch->voice; vc; vc = vc->nxt) {
         fprintf(out, "  %d ms", vc->time);
         if (vc->is_off) fprintf(out, " (OFF)");
         if (!vc->vtyp) {
            fprintf(out, " NO VTYP\n");
         } else {
            vc->vtyp->dump(vc, out);
         }
      }
   }
}

// END //
